;; logging.scm -- output and log utilities

;; accumulate boot messages until syslog can ingest them
(define *message-buffer* '())

(define (boot-log str)
  "Writes str to /dev/log"
  (catch-system-error
   (let ((sock (socket PF_UNIX SOCK_DGRAM 0)))
     (sendto sock
	     (string->bytevector str "UTF-8")
	     AF_UNIX "/dev/log")
     (close-port sock))))

;; if we don't delay *tput-start-ok* initialization, guile 2.0 with
;; shepherd 0.3.1 crashes with SIGCHLD on waitpid -1
(define *tput-start-ok* #f)

(define (build-tput-start-ok)
  (set! *tput-start-ok* (cat (system-value* "/usr/bin/tput" "civis")
			     (system-value* "/usr/bin/tput" "sc")
			     (system-value* "/usr/bin/tput" "hpa" "1")
			     (system-value* "/usr/bin/tput" "setaf" "2")
			     " Ok "
			     (system-value* "/usr/bin/tput" "rc")
			     (system-value* "/usr/bin/tput" "cnorm")
			     (system-value* "/usr/bin/tput" "op")
			     "\n"))
  )

;; ditto here
(define *tput-start-fail* #f)

(define (build-tput-start-fail)
  (set! *tput-start-fail* (cat (system-value* "/usr/bin/tput" "civis")
			       (system-value* "/usr/bin/tput" "sc")
			       (system-value* "/usr/bin/tput" "hpa" "1")
			       (system-value* "/usr/bin/tput" "setaf" "1")
			       "FAIL"
			       (system-value* "/usr/bin/tput" "rc")
			       (system-value* "/usr/bin/tput" "cnorm")
			       (system-value* "/usr/bin/tput" "op")
			       "\n"))
  )

;; display doesn't work at boot in shepherd, so we go around that;
;; notice that the proc file system must be mounted, which is
;; normally done in initramfs

(define *s* (open-output-file "/proc/self/fd/1"))

(define (boot-format-ok)
  (unless *tput-start-ok* (build-tput-start-ok))
  (simple-format *s* *tput-start-ok*)
  (force-output *s*))

(define (boot-format-fail)
  (unless *tput-start-fail* (build-tput-start-fail))
  (simple-format *s* *tput-start-fail*)
  (force-output *s*))

(define (uptime)
  "Get uptime since boot in seconds, formatted like real kernel messages"
  (let* ((s (open-input-file "/proc/uptime"))
	 (val (read s)))
    (close s)
    (let* ((str (simple-format #f "~a" val))
	   (spl (string-split str #\.))
	   (a (car spl))
	   (b (cadr spl)))
      (string-concatenate (list (make-string (- 5 (string-length a)) #\sp) a "."
				b (make-string (- 6 (string-length b)) #\sp))))))

(define boot-format
    (lambda (format . args)
      ;; we write to console anyway
      (apply simple-format *s* format args)
      (force-output *s*)

      (when (and (not (file-exists? "/dev/log"))
		 (file-exists? "/dev/kmsg")
		 (file-exists? "/bin/dmesg"))
	;; we read kernel boot messages directly until rsyslog takes
	;; over, so as to interlace them properly with shepherd messages
	(system-pipe "/bin/dmesg -c"
		     (lambda (str)
		       (set! *message-buffer* (cons (simple-format #f "<6>~a" str) *message-buffer*)))))

      ;; as rsyslogd by default uses /proc/kmsg through imklog, we
      ;; need to empty that data source to avoid duplicate messages
      (when (and (not (file-exists? "/dev/log"))
      		 (file-exists? "/proc/kmsg"))
      	(let ((s (open "/proc/kmsg" O_RDONLY O_NDELAY)))
      	  (do ()
      	      ((not (char-ready? s)) (close s))
      	    (read-line s))))

      ;; we write to /dev/log if available, that means rsyslog has started running
      ;; the first time we get here, we have, outside the current message,
      ;; 3 sets of data on hold competing to get rsyslog attention:
      ;;   * the kernel kmsg ring buffer
      ;;   * shepherd itself (messages like "Service xxx has been started...")
      ;;   * our not empty buffer *message-buffer*
      (if (file-exists? "/dev/log")
	  (begin
	    (unless (null? *message-buffer*)
	      (for-each boot-log
			(reverse *message-buffer*))
	      (set! *message-buffer* '()))
	    ;; 30 = 8*3 + 6 (3: daemon; 6: info)
	    (boot-log (simple-format #f "<30>~a shepherd[1]: ~a"
				     (strftime "%b %d %H:%M:%S" (localtime (current-time)))
				     (apply simple-format #f format args))))
	  ;; if not, bufferize the message
          (set! *message-buffer* (cons (simple-format #f "<6>[~a] shepherd[1]: ~a"
						      (uptime)
                                                      (apply simple-format #f format args))
				       *message-buffer*)))
      ))
