;; mount.scm -- mount and umount in scheme

(use-modules (system foreign))

(define MS_RDONLY 1)
(define MS_NOSUID 2)
(define MS_NODEV 4)
(define MS_NOEXEC 8)
(define MS_REMOUNT 32)
(define MS_NOATIME 1024)
(define MS_NODIRATIME 2048)
(define MS_RELATIME (expt 2 21))
(define MS_NOUSER (expt 2 31))

(define mount-raw
  (let ((this (dynamic-link)))
    (pointer->procedure int
                        (dynamic-func "mount" this)
                        (list '* '* '* unsigned-long '*)
			#:return-errno? #t)))

(define (mount dev node type options options2)
  "Scheme version of mount(2)"
  (mount-raw (string->pointer dev)
             (string->pointer node)
             (string->pointer type)
	     options (string->pointer options2)))

(define (is-mountpoint? str)
  "Test if str exists and is the path of a mounted filesystem"
  (or (string= "/" str)
      (and (file-exists? str)
	   (not (= (stat:dev (stat str)) (stat:dev (stat (cat str "/.."))))))))

(define umount-raw
  (let ((this (dynamic-link)))
    (pointer->procedure int
                        (dynamic-func "umount" this)
                        (list '*)
			#:return-errno? #t)))

(define (umount path)
  "Scheme version of umount(2)"
  (umount-raw (string->pointer path)))

(define MNT_DETACH 2)

(define umount2-raw
  (let ((this (dynamic-link)))
    (pointer->procedure int
                        (dynamic-func "umount2" this)
                        (list '* int)
			#:return-errno? #t)))

(define (umount2 path opts)
  "Scheme version of umount2(2)"
  (umount2-raw (string->pointer path) opts))

(use-modules (ice-9 rdelim))
(use-modules (ice-9 regex))

(define (rootfs)
  "Get characteristics of / filesystem in /etc/fstab"
  (let ((s (open-input-file "/etc/fstab"))
	(devid #f))
    (do ()
	((or (port-closed? s) (eof-object? (peek-char s))) (close s) devid)
      (let* ((str (read-line s))	     
	     (strings (string-split (regexp-substitute/global #f "[ \t]+" str 'pre " " 'post) #\Space)))
	;;(simple-format #t "~s ~a\n" strings (length strings))
	(when (and (not (char=? #\# (string-ref str 0)))
		   (= (length strings) 6))
	  (when (string= (list-ref strings 1) "/")
	    (set! devid (list (list-ref strings 0) (list-ref strings 2) (list-ref strings 3)))
	    (close s)))))))

(define (rootfs-proc)
  "Get characteristics of / filesystem in /proc/mounts"
  (let ((s (open-input-file "/proc/mounts"))
	(devid #f))
    (do ()
	((or (port-closed? s) (eof-object? (peek-char s))) (close s) devid)
      (let* ((str (read-line s))	     
	     (strings (string-split (regexp-substitute/global #f "[ \t]+" str 'pre " " 'post) #\Space)))
	;;(simple-format #t "~s ~a\n" strings (length strings))
	(when (and (not (char=? #\# (string-ref str 0)))
		   (= (length strings) 6))
	  (when (string= (list-ref strings 1) "/")
	    (set! devid (list (list-ref strings 0) (list-ref strings 2) (list-ref strings 3)))
	    (close s)))))))

(define (rootfs-mntid)
  "Get mount id of / filesystem from /proc/self/mountinfo"
  (let ((s (open-input-file "/proc/self/mountinfo"))
	(devid #f))
    (do ()
	((or (port-closed? s) (eof-object? (peek-char s))) (close s) devid)
      (let* ((str (read-line s))	     
	     (strings (string-split str #\Space)))
	;;(simple-format #t "~s ~a\n" strings (length strings))
	(when (string= (list-ref strings 4) "/")
	  (set! devid (string->number (list-ref strings 0)))
	  (close s))))))

(define (get-mounted-fs)
  "Get mounted filesystem from /proc/mounts candidates for unmounting, except / and /usr"
  (let ((s (open-input-file "/proc/mounts"))
	(fs '()))
    (do ()
        ((or (port-closed? s) (eof-object? (peek-char s))) (close s) fs)
      (let* ((str (read-line s))
             (strings (string-split (regexp-substitute/global #f "[ \t]+" str 'pre " " 'post) #\Space)))
        ;;(simple-format #t "~s ~a\n" strings (length strings))
        (when (and (not (char=? #\# (string-ref str 0)))
                   (= (length strings) 6))
          (unless (string-match "^(/|/usr|/proc|/dev|/\\.dev|/dev/pts|/dev/shm|/dev/\\.static/dev|/proc/.*|/sys|/sys/.*|/run|/run/.*|/dev/vcs)$" (list-ref strings 1))
            (set! fs (cons (list (list-ref strings 0) (list-ref strings 1) (list-ref strings 2)) fs))))))
    ))

(define (get-mounted-usr)
  "Get /usr filesystem details from /proc/mounts"
  (let ((s (open-input-file "/proc/mounts"))
	(fs #f))
    (do ()
        ((or (port-closed? s) (eof-object? (peek-char s))) (close s) fs)
      (let* ((str (read-line s))
             (strings (string-split (regexp-substitute/global #f "[ \t]+" str 'pre " " 'post) #\Space)))
        ;;(simple-format #t "~s ~a\n" strings (length strings))
        (when (and (not (char=? #\# (string-ref str 0)))
                   (= (length strings) 6))
          (when (string-match "^/usr$" (list-ref strings 1))
            (set! fs (list (list-ref strings 0) (list-ref strings 1) (list-ref strings 2)))))))
    ))


