;; various.scm -- various helper functions

(define (create-static-dev osrelease)
  "Parses modules.devname to create static devices"
  (let ((s (open-input-file (cat "/lib/modules/" osrelease "/modules.devname"))))
    (do ()
	((eof-object? (peek-char s)))
      (let* ((str (read-line s))
	     (strings (string-split str #\Space)))
	(when (= 3 (length strings))
	  (let* ((devcodes (string-split (list-ref strings 2) #\:))
		 (type (substring (car devcodes) 0 1))
		 (ucode (substring (car devcodes) 1))
		 (gcode (cadr devcodes))
		 (dev (+ (* (with-input-from-string ucode read) 256) (with-input-from-string gcode read)))
		 (path (string-split (cadr strings) #\/)))
	    (when (< 1 (length path))
	      (let ((dir (dirname (cadr strings))))
		(when (not (file-exists? (string-append "/dev/" dir)))
		  (mkdir (cat "/dev/" dir)))))
	    (catch-system-error  ;; in case the node already exists
	     (mknod (cat "/dev/" (cadr strings)) (if (string= type "c") 'char-special 'block-special) #o600 dev))
	    ;;(boot-format "mknod -m 0600 /dev/~a ~a ~a ~a\n" (cadr strings) type ucode gcode)
	    ))))))

(define (scan dir)
  "Recursively deletes content of directory dir"
  (when (and (file-exists? dir) (eq? (stat:type (lstat dir)) 'directory))
    (for-each (lambda (x)
		(when (and (not (string= "." x))
			   (not (string= ".." x))
			   ;; in case dir is the root of a mounted filesystem, we skip it
			   (not (and (is-mountpoint? dir)
				     (string= "lost+found" x))))
		  (if (eq? (stat:type (lstat (string-append dir "/" x))) 'directory)
		      (unless (is-mountpoint? (string-append dir "/" x))
			;; so we don't cross filesystems
			(scan (string-append dir "/" x))
			;; will fail is directory not empty
			;; (e.g. because there is a mount point below)
			(catch 'system-error
			       (lambda ()
				 ;;(boot-format "rmdir ~a/~a\n" dir x)
				 (rmdir (string-append dir "/" x)))
			       (lambda (keys . args) #f)))
		      (begin
			;;(boot-format "~a/~a\n" dir x)
			(delete-file (string-append dir "/" x))))))
	      (scandir dir))))

(define (clean-all)
  "Cleaning /tmp"
  ;; as /run, /run/lock and /run/shm are of tmpfs type, no
  ;; need to clean them; we only clean /tmp
  (scan "/tmp")
  )

(define (get-random-poolsize)
  "Get size of random seed file as a string"
  (if (file-exists? "/proc/sys/kernel/random/poolsize")
      (with-input-from-string
	  (let* ((s (open-input-file "/proc/sys/kernel/random/poolsize"))
                 (name (read-delimited "\n" s)))
            (close s)
	    name)
	;;(system-value "cat /proc/sys/kernel/random/poolsize 2>/dev/null")
	(lambda ()
	  (with-output-to-string (lambda () (write (quotient (+ 7 (read)) 8))))))
      "512"))

;; ECHILD error can happen
(define (waitpid* pid options)
  "waitpid with clean handling of ECHILD error"
  (catch 'system-error
	 (lambda ()
	   (waitpid pid options))
	 (lambda (keys . args)
	   (cons 0 0))))

;; sleep doesn't work on shepherd 0.10.1 and above
(define (try-every-second n func)
  "Execute func every second until success; fails if more than n
iterations"
  (let ((i 1)
	(old (sigaction SIGALRM)))
    (define (prox)
      (sigaction SIGALRM (lambda (x) (set! i (+ 1 i))))
      (alarm 1)
      (pause)
      (let ((ret (func)))
	(when (and (< i n) (not ret)) (prox))))
    (prox)
    (sigaction SIGALRM (car old) (cdr old))
    (if (>= i n)
	#f
	#t)))

