;; shepherd.scm -- default shepherd configuration file.

(define cat string-append)

(use-modules (ice-9 popen))
(use-modules (ice-9 rdelim))
(use-modules (ice-9 ftw))
(use-modules (ice-9 regex))
(use-modules (ice-9 receive))
(use-modules (ice-9 iconv))
(use-modules (shepherd support))
(use-modules (shepherd config))

(load (cat %sysconfdir "/shepherd.d/utils/processes.scm"))
(load (cat %sysconfdir "/shepherd.d/utils/logging.scm"))
(load (cat %sysconfdir "/shepherd.d/utils/mount.scm"))
(load (cat %sysconfdir "/shepherd.d/utils/fuser.scm"))
(load (cat %sysconfdir "/shepherd.d/utils/utmp.scm"))
(load (cat %sysconfdir "/shepherd.d/utils/various.scm"))

;; from kernel boot parameters
(define debug (member "debug_shepherd=y" (environ)))

;; we can revert to (guile-user) while shepherd runs, so keep track of that
(define shepherd-module (current-module))

;; Services known to shepherd:
;; Add new services (defined using 'make <service>') to shepherd here by
;; providing them as arguments to 'register-services'.
(register-services
 (make <service>
    #:requires '(root)
    #:provides '(hostname-setup)
    #:start (lambda args
 	      (boot-format " -- Booting --\n")
	      (boot-format "[....] Setting hostname...")
	      (if (file-exists? "/etc/hostname")
		  (let* ((s (open-input-file "/etc/hostname"))
			 (name (read-delimited "\n" s)))
		    (close s)
		    (sethostname name)
		    (boot-format-ok))
		  (boot-format-fail))
	      #t
	      )
    #:stop (lambda (running)
 	     (boot-format " -- Ready to reboot or shut down --\n")
	     ;; so we can analyze halt/reboot failures
	     (when debug (system "/bin/sh"))
	     #f)
    )
  (make <service>
    #:requires '(hostname-setup)
    #:provides '(mount-in-memory-fs)
    #:start (lambda args
	      (when (file-exists? "/run")
		(unless (is-mountpoint? "/run")
		  ;; normally mounted in initrd
                  (boot-format "[....] Mounting /run...")
		  (receive (ret err)
		      (mount "tmpfs" "/run" "tmpfs" (+ MS_NOSUID MS_NOEXEC) "")
		    (if (zero? ret)
			(boot-format-ok)
			(begin
			 (boot-format "~a" (strerror err))
			 (boot-format-fail)
			 )))))

	      ;; the processes with their pid files linked there won't
	      ;; be killed early at reboot/shutdown (ex: rsyslogd)
	      (unless (file-exists? "/run/sendsigs.omit.d")
		(mkdir "/run/sendsigs.omit.d"))
	      (chmod "/run/sendsigs.omit.d" #o755)
		    
	      (unless (file-exists? "/run/lock")
		(mkdir "/run/lock"))
	      (chmod "/run/lock" #o755)		    
	      (unless (is-mountpoint? "/run/lock")
		(boot-format "[....] Mounting /run/lock...")
		(receive (ret err)
		    (mount "tmpfs" "/run/lock" "tmpfs" (+ MS_NODEV MS_NOSUID MS_NOEXEC) "")
		  (if (zero? ret)
		      (boot-format-ok)
		      (boot-format "failed: ~a\n" (strerror err)))))
	      
	      (when (file-exists? "/proc")
		(unless (is-mountpoint? "/proc")
                  (boot-format "Mounting /proc...")
		  ;; normally mounted in initrd
		  (receive (ret err)
		      (mount "proc" "/proc" "proc" (+ MS_NODEV MS_NOSUID MS_NOEXEC) "")
		    (if (zero? ret)
			(boot-format-ok)
			(boot-format "failed: ~a\n" (strerror err))))))

	      (when (file-exists? "/sys")
		(unless (is-mountpoint? "/sys")
                  (boot-format "[....] Mounting /sys...")
		  ;; normally mounted in initrd
		  (receive (ret err)
		      (mount "sysfs" "/sys" "sysfs" (+ MS_NODEV MS_NOSUID MS_NOEXEC) "")
		    (if (zero? ret)
			(boot-format-ok)
			(boot-format "failed: ~a\n" (strerror err))))))

	      (when (file-exists? "/sys/fs/pstore")
		(unless (is-mountpoint? "/sys/fs/pstore")
                  (boot-format "[....] Mounting /sys/fs/pstore...")
		  (receive (ret err)
		      (mount "pstore" "/sys/fs/pstore" "pstore" MS_RELATIME "")
		    (if (zero? ret)
			(boot-format-ok)
			(boot-format "failed: ~a\n" (strerror err))))))
	      
	      (when (file-exists? "/sys/kernel/config")
		(unless (is-mountpoint? "/sys/kernel/config")
                  (boot-format "[....] Mounting /sys/kernel/config...")
		  (receive (ret err)
		      (mount "configfs" "/sys/kernel/config" "configfs" 0 "")
		    (if (zero? ret)
			(boot-format-ok)
			(boot-format "failed: ~a\n" (strerror err))))))
	      
	      #t
	      )
    )
  (make <service>
    #:requires '(mount-in-memory-fs)
    #:provides '(udev)
    #:start (lambda args
	      ;; remount /dev; for this, we need to first unmount
	      ;; /dev/pts and /dev/shm (they will be remounted later)
	      (when (is-mountpoint? "/dev/pts")
		(boot-format "[....] Unmounting /dev/pts...")
		(receive (ret err)
		    (umount2 "/dev/pts" MNT_DETACH)
		  (if (zero? ret)
		      (boot-format-ok)
		      (boot-format "failed: ~a\n" (strerror err)))))

	      (when (is-mountpoint? "/dev/shm")
		(boot-format "[....] Unmounting /dev/shm...")
		(receive (ret err)
		    (umount2 "/dev/shm" MNT_DETACH)
		  (if (zero? ret)
		      (boot-format-ok)
		      (boot-format "failed: ~a\n" (strerror err)))))
 		
	      (when (is-mountpoint? "/dev")
		(boot-format "[....] Remounting /dev...")
		(receive (ret err)
		    (mount "devtmpfs" "/dev" "devtmpfs" MS_REMOUNT "size=10M,mode=0755")
		  (if (zero? ret)
		      (boot-format-ok)
		      (boot-format "failed: ~a\n" (strerror err)))))
 		
	      ;; creation of devices nodes in /dev already known by
	      ;; kernel; only new nodes will be handled by udev
	      (if (file-exists? "/proc/sys/kernel/osrelease")
		  (let* ((s (open-input-file "/proc/sys/kernel/osrelease"))
			 (str (read-delimited "\n" s)))
		    (close s)
		    (when (file-exists? (cat "/lib/modules/" str "/modules.devname"))
		      (boot-format "[....] Create static dev nodes...")
		      (create-static-dev str)
		      (boot-format-ok))
		    ))
	      
	      (when (file-exists? "/lib/systemd/systemd-udevd") 
		(boot-format "[....] Starting udev...")
		(system* "/bin/udevadm" "info" "--cleanup-db")
	      	(system "/lib/systemd/systemd-udevd --daemon 2>/dev/null")
		(boot-format "version ~a..." (system-value* "udevadm" "info" "--version"))
		;; should we create a pid file for udev?
		(system* "/bin/udevadm" "trigger" "--action=add")
		(system* "/bin/udevadm" "settle")
		(boot-format-ok))

	      ;; still useful?
	      (if (file-exists? "/sbin/MAKEDEV")
		  (symlink "/sbin/MAKEDEV" "/dev/MAKEDEV")
		  (symlink "/bin/true" "/dev/MAKEDEV"))
	  
	      #t
	      )
    #:stop (lambda (running)
    	     (when running
		   (boot-format "[....] Stopping udev...")
		   (catch #t
			  (lambda ()
			    ;; it can happen that the following crashes, so kill directly
			    ;; (system "udevadm control --exit")
			    (let ((pid (system-value-read "pgrep -d ' ' -u root systemd-udevd")))
			      (when (exact-integer? pid)
				(kill pid SIGKILL))
			      )

			    (boot-format-ok))
			  (lambda (keys . args) (boot-format-fail)))
		   )
	     
	     ;; stopping a service often involves killing it by a
	     ;; signal with no handler to catch it and exit
	     ;; properly; so we end up with zombies we have to
	     ;; delete; do that first
             (boot-format "[....] Deleting zombies...")
	     (do ((i 1 (1+ i)))
		 ((let ((ret (waitpid* -1 WNOHANG)))
		    (zero? (car ret)))))
	     (boot-format-ok)
	     
	     (let ((services (scandir "/run/sendsigs.omit.d" (lambda (x) (string-match ".*" x))))
		   (exceptions '()))
	       (for-each (lambda (x)
			   (unless (or (string= x ".") (string= x ".."))
			     (catch 'system-error
				    (lambda ()
				      (set! exceptions
					    (cons (with-input-from-file (cat "/run/sendsigs.omit.d/" x) read)
						  exceptions)))
				    (lambda (keys . args) (boot-format "Reading /run/sendsigs.omit.d/~a failed\n" x)))))
			 services)

	       ;; delete real processes not within exceptions
               (boot-format "[....] Asking all remaining processes to terminate...")
	       (apply killall5 SIGTERM exceptions)
	       (boot-format-ok)

	       ;; restart stopped processes
	       (boot-format "[....] Waiting processes to complete...")
	       (apply killall5 SIGCONT exceptions)
	       (boot-format-ok)
	       
	       ;; killing by force remaining processes
	       (boot-format "[....] Killing all remaining processes...")
	       (apply killall5 SIGKILL exceptions)
	       (boot-format-ok))

	     ;; and delete them if they end up as zombies 
             (boot-format "[....] Deleting last zombies...")
	     (do ((i 1 (1+ i)))
		 ((let ((ret (waitpid* -1 WNOHANG)))
		    (zero? (car ret)))))
	     (boot-format-ok)

    	     #f)
    )
  (make <service>
    #:requires '(udev)
    #:provides '(keyboard-setup)
    #:start (lambda args
	      (boot-format "[....] Setting up keyboard...")
	      (when (file-exists? "/bin/setupcon")
 		(system "/bin/setupcon -k 2>/dev/null"))
	      (boot-format-ok)
	      #t
	      )
    #:stop (lambda (running)
	     (when debug (system "/bin/sh"))
	     #f)
    )
  (make <service>
    #:requires '(keyboard-setup)
    #:provides '(mount-in-memory-fs-step2)
    #:start (lambda args
	      (unless (file-exists? "/run/shm")
		(mkdir "/run/shm")
		(chmod "/run/shm" #o755))
	      (unless (is-mountpoint? "/run/shm")
		(boot-format "[....] Mounting /run/shm...")
		(receive (ret err)
		    (mount "tmpfs" "/run/shm" "tmpfs" (+ MS_NODEV MS_NOSUID MS_NOEXEC) "")
		  (if (zero? ret)
		      (boot-format-ok)
		      (boot-format "failed: ~a\n" (strerror err)))))

	      ;; normally, we have:
	      ;; ln -sf /run /var/run
	      ;; ln -sf /run/lock /var/lock
	      ;; ln -sf /run/shm /dev/shm
	      ;; the 2 first are done at installation,
	      ;; the last one must be redone at every boot
	      (symlink "/run/shm" "/dev/shm")
	      
	      (unless (file-exists? "/dev/pts")
		(mkdir "/dev/pts")
		(chmod "/dev/pts" #o755))
	      (unless (is-mountpoint? "/dev/pts")
		(boot-format "[....] Mounting /dev/pts...")
		(receive (ret err)
		    (mount "devpts" "/dev/pts" "devpts" (+ MS_NOSUID MS_NOEXEC) "gid=5,mode=620")
		  (if (zero? ret)
		      (boot-format-ok)
		      (boot-format "failed: ~a\n" (strerror err)))))
	      
	      #t
	      )
    )
   (make <service>
    #:requires '(mount-in-memory-fs-step2)
    #:provides '(clock-setup)
    #:start (lambda args
	      (boot-format "[....] Setting clock...")
 	      ;; If the admin deleted the hwclock config, create a blank
	      ;; template with the defaults.
	      (when (and (logbit? 7 (stat:perms (stat "/etc"))) (not (file-exists? "/etc/adjtime")))
		(boot-format "creating /etc/adjtime...")
		;; hmmm... at that step, the root file system is still
		;; mounted ro, so we put /etc/adjtime elsewhere
		(with-output-to-file "/run/adjtime" (lambda () (simple-format #t "0.0 0 0.0\n0\nUTC\n"))))
	      
	      (system* "/sbin/hwclock" "--adjfile=/run/adjtime" "--rtc=/dev/rtc0" "--hctosys")
	      (boot-format-ok)
	      #t
	      )
    ;; #: stop action moved just before remounting / ro 
    )
   (make <service>
    #:requires '(clock-setup)
    #:provides '(lvm2)
    #:start (lambda args
	      (when (file-exists? "/sbin/vgchange")
		(boot-format "[....] Setting up LVM Volume Groups...")
		(system "LVM_SUPPRESS_FD_WARNINGS= /sbin/lvm vgchange -aay --sysinit >/dev/null")
		(boot-format-ok))
	      #t
	      )
    )
   (make <service>
    #:requires '(lvm2)
    #:provides '(checkroot)
    #:start (lambda args
	      (system "/sbin/swapon -a -e >/dev/null")

	      ;; devid is like UUID=... and is read in /etc/fstab; it
	      ;; is normally the same as root= on the kernel command
	      ;; line

	      ;; rootfs is type of root filesystem
	      ;; rootopts are mount options of filesystem
	      ;; rootdev is like /dev/sda1

	      ;; ddev and rdev should be equal
	      (let* ((rootfs-details (rootfs))
		     (devid (list-ref rootfs-details 0)) ;(system-value "/bin/findmnt -s -n -o SOURCE /"))
		     (rootfs (list-ref rootfs-details 1)) ;(system-value "/bin/findmnt -s -n -o FSTYPE /"))
		     (rootopts (list-ref rootfs-details 2)) ;(system-value "/bin/findmnt -s -n -o OPTIONS /"))
		     (rootdev (system-value* "/sbin/findfs" devid))
		     (ddev (system-value* "/bin/mountpoint" "-qx" rootdev))
		     (rdev (system-value* "/bin/mountpoint" "-d" "/")))
		;;(boot-format "~a\n" rootfs-details)
		(when (not (string= ddev rdev))
		  ;; in original init-sysv, we create in that case rdev on /run/rootdev and remount it on /
		  (system* "/bin/echo" (cat "Warning: cannot find rootdev rdev=" rdev " ddev=" ddev)))
		
		;; root has already been fscked in initramfs 
		(system (cat "/sbin/fsck -a -t " rootfs " " rootdev "; FSCKCODE=$?; if [ $FSCKCODE -gt 3 ]; then /sbin/sulogin; fi"))

		;; remount rw
		(boot-format "[....] Remounting / rw...")
		(system (cat "mount -n -o remount," rootopts ",rw " devid " / 2>/dev/null || mount -n -o remount," rootopts ",rw /")))
		(boot-format-ok)

		;; move /run/adjtime at the right place
		(when (file-exists? "/run/adjtime")
		  (copy-file "/run/adjtime" "/etc/adjtime")
		  (delete-file "/run/adjtime"))
		
	      (when (file-exists? "/proc/mounts")
		(unless (and (file-exists? "/etc/mtab")
			     (eq? 'symlink (stat:type (lstat "/etc/mtab")))
			     (string= "/proc/mounts" (readlink "/etc/mtab")))
		  (delete-file "/etc/mtab")
		  (symlink "/proc/mounts" "/etc/mtab")))
	      #t
	      )
    #:stop (lambda (running)
	     ;; we have already unmounted everything else on disk,
	     ;; except / and /usr; this last one because guile resides
	     ;; there, so we remount it ro if it is a separate
	     ;; partition
	     (let ((usrfs-details (get-mounted-usr)))
	       (when usrfs-details
		 (let ((devid (list-ref usrfs-details 0))
		       (userfs (list-ref usrfs-details 1)))
		   (boot-format "[....] Remounting /usr (~a) read only..." devid)
		   (receive (ret err)
		       (mount devid "/usr" userfs (+ MS_REMOUNT MS_RDONLY) "")
		     (if (zero? ret)
			 (boot-format-ok)
			 (boot-format "failed: ~a\n" (strerror err))))))
	       )

	     ;; as we write to /etc/adjtime, must be done just before remounting / ro
	     (when running
		   (boot-format "[....] Saving clock...")
		   (system* "/sbin/hwclock" "--rtc=/dev/rtc0" "--systohc")
		   (boot-format-ok)
		   )

	     ;; remount / as ro
	     (let* ((rootfs-details (rootfs-proc))
		    (devid (list-ref rootfs-details 0))
		    (rootfs (list-ref rootfs-details 1)))
	       (let ((pids (fuser)))
		 (for-each (lambda (x) (if (= x 1)
					   (boot-format "shepherd has a file on / open writing, should not happen...")
					   (catch 'system-error
						  (lambda () (kill x SIGTERM) (boot-format "kill ~a SIGTERM ok\n" x))
						  (lambda (key . args) (boot-format "kill ~a SIGTERM failed\n" x)))))
			   pids)
		 (for-each (lambda (x) (if (= x 1)
                                           (boot-format "shepherd has a file on / open writing, should not happen...")
					   (catch 'system-error
						  (lambda () (kill x SIGKILL) (boot-format "kill ~a SIGKILL ok\n" x))
						  (lambda (key . args) (boot-format "kill ~a SIGKILL failed\n" x)))))
			   pids)
		 )
               (boot-format "[....] Remounting / (~a) read only..." devid)
	       (receive (ret err)
		   (mount devid "/" rootfs (+ MS_REMOUNT MS_RDONLY) "")
		 (if (zero? ret)
		     (boot-format-ok)
		     (boot-format "failed: ~a\n" (strerror err)))))
	     #f)
    )
   (make <service>
    #:requires '(checkroot)
    #:provides '(checkfs)
    #:start (lambda args
	      ;; /usr if it exists has already been fscked and mounted in initramfs, hence -M
	      (system* "/sbin/fsck" "-V" "-M" "-R" "-A" "-a")
	      #t
	      )
    )
   (make <service>
    #:requires '(checkfs)
    #:provides '(checkroot-bootclean)
    #:start (lambda args
	      (boot-format "[....] Cleaning up temporary files...")
	      ;; at that point, if /tmp is a separate filesystem, it is not mounted
	      (clean-all)
	      (boot-format-ok)
	      #t
	      )
    )
   (make <service>
    #:requires '(checkroot-bootclean)
    #:provides '(load-kernel-modules)
    #:start (lambda args
	      (let ((we-want-etc-modules #t)
		    (modules-files '()))
		(for-each (lambda (dir)
			    (when (file-exists? dir)
			      (boot-format "[....] Looking for modules files in ~a..." dir)
			      
			      (let ((files (scandir dir (lambda (x) (string-match ".*\\.conf$" x)))))
				(for-each (lambda (file)
					    (let ((fullname (cat dir "/" file)))
					      (boot-format "Found file ~a... " fullname)
			      		    
					      ;; we drop duplicates
					      (unless (member fullname modules-files)
						;; we add a symlink to
						;; /etc/modules only one
						;; time
						(when (and we-want-etc-modules
							   (eq? 'symlink (stat:type (lstat fullname)))
							   (string= "/etc/modules" (system-value* "realpath" fullname)))
						  (boot-format "Got /etc/modules as ~a..." fullname)
						  (set! we-want-etc-modules #f))
						
						(set! modules-files (cons (system-value* "realpath" fullname) modules-files)))))
					  files))
			      (boot-format-ok)))
			  (list "/etc/modules-load.d" "/run/modules-load.d" "/lib/modules-load.d"))

		;; we add /etc/modules anyway, if not already there
		(when we-want-etc-modules
		      (boot-format "[....] Adding /etc/modules")
		      (set! modules-files (cons "/etc/modules" modules-files))
		      (boot-format-ok))

		(boot-format "[....] Really doing modprobe now...")
		(for-each (lambda (file)
			    (boot-format "Loading kernel modules file ~a..." file)
			    (system-pipe (cat "cat " file)
					 (lambda (x)
					   (when (> (string-length x) 0)
					     (unless (string-match "#.*" x)
					       (boot-format "Loading kernel module ~a..." x)
					       (system* "modprobe" x)))))
			    )
			  (reverse modules-files))
		(boot-format-ok))
		
	      #t
	      )
    )
   (make <service>
    #:requires '(load-kernel-modules)
    #:provides '(mountall)
    #:start (lambda args
	      ;; if /usr is already mounted, force it rw
	      (let ((usrfs-details (get-mounted-usr)))
		(when usrfs-details
		      (let ((devid (list-ref usrfs-details 0))
			    (userfs (list-ref usrfs-details 1)))
			(boot-format "[....] Remounting /usr (~a) read write..." devid)
			(receive (ret err)
				 (mount devid "/usr" userfs MS_REMOUNT "")
				 (if (zero? ret)
				     (boot-format-ok)
				     (boot-format "failed: ~a\n" (strerror err))))))
		)

	      ;; mount the fs on disk
	      (boot-format "[....] Mounting remaining filesystems on disk...")
	      (system* "mount" "-a" "-t" "nonfs,nfs4,smbfs,cifs,ncp,ncpfs,coda,ocfs2,gfs,gfs2,ceph" "-O" "no_netdev")
	      ;; post_mountall
	      (boot-format-ok)
	      #t
	      )
    #:stop (lambda (running)
	      ;; at that point, if /tmp is a separate filesystem, it is mounted
	      (for-each (lambda (fs)
			  (boot-format "[....] Unmounting filesystem ~a..." (list-ref fs 1))
			  (receive (ret err)
			      (umount2 (list-ref fs 1) MNT_DETACH)
			    (if (zero? ret)
				(boot-format-ok)
				(begin
				  (boot-format "~a" (strerror err))
				  (boot-format-fail)
				  )))
			  )
			(get-mounted-fs))
	      #f
	      )
    )
   (make <service>
    #:requires '(mountall)
    #:provides '(mountall-bootclean)
    #:start (lambda args
	      (boot-format "[....] Cleaning up temporary files...")
	      ;; at that point, if /tmp is a separate filesystem, it is mounted
	      (clean-all)
	      (boot-format-ok)
	      #t
	      )
    )
   (make <service>
    #:requires '(mountall-bootclean)
    #:provides '(kernel-variables-setup)
    #:start (lambda args
	      (boot-format "[....] Setting kernel variables...")
	      (system* "/sbin/sysctl" "-q" "--system")
	      (boot-format-ok)
	      #t
	      )
    )
   (make <service>
    #:requires '(kernel-variables-setup)
    #:provides '(random-generator-setup)
    #:start (lambda args
	      (boot-format "[....] Initializing random number generator...")
	      (system "(date +%s.%N; if [ -f /var/lib/urandom/random-seed ]; then cat /var/lib/urandom/random-seed; fi) >/dev/urandom")
	      (let ((poolsize (get-random-poolsize)))
		;; saved in case of crash before orderly shutdown
		(system (cat "dd if=/dev/urandom of=/var/lib/urandom/random-seed bs=" poolsize " count=1 >/dev/null 2>&1"))
		(boot-format-ok))
	      #t
	      )
    #:stop (lambda (running)
	     (boot-format "[....] Saving random number seed...")
	     (let ((poolsize (get-random-poolsize)))
		(system (cat "dd if=/dev/urandom of=/var/lib/urandom/random-seed bs=" poolsize " count=1 >/dev/null 2>&1"))
		(boot-format-ok))
	      #f
	      )
    )
   (make <service>
    #:requires '(random-generator-setup)
    #:provides '(networking)
    #:start (lambda args
	      (boot-format "[....] Configuring network interfaces...")
	      (unless (file-exists? "/run/network")
		(mkdir "/run/network")
		(let ((gid (catch-system-error (group:gid (getgrnam "netdev")))))
		  (when gid (chown "/run/network" -1 gid))))
	      
	      (when (file-exists? "/sbin/udevadm")
		(let ((str ""))
		  (system-pipe "/sbin/ifquery --list --exclude=lo" (lambda (s) (set! str (string-append str s))))
                  (system-pipe "/sbin/ifquery --list --allow hotplug" (lambda (s) (set! str (string-append str s))))
 		  (when (> (string-length str) 0)
		    (system* "/sbin/udevadm" "settle"))))
	      
	      (system* "/sbin/ifup" "-a")

	      ;; managing hotplug seems a bit brittle
	      (when (file-exists? "/sys/class/net")
		(system-pipe "/sbin/ifquery --list --allow hotplug"
			     (lambda (iface)
			       (when (file-exists? (cat "/sys/class/net/" iface))
				 (system* "/sbin/ip" "link" "set" iface "up")
				 (when (string= "up" (system-value* "cat" (cat "/sys/class/net/" iface "/operstate")))
				   (boot-format "~a..." iface))
				 )))
		  
		(system-pipe "/sbin/ifquery --list --allow hotplug"
			     (lambda (iface)
			       (system (cat "/sbin/ifquery --state " iface " >/dev/null || /sbin/ifup " iface))))
		)
	      (boot-format-ok)
	      #t
	      )
    #:stop (lambda (running)
	     (boot-format "[....] Unconfiguring network interfaces...")
	     (catch 'system-error
		    (lambda ()
		      (system* "/sbin/ifdown" "-a" "--exclude=lo")
		      (boot-format-ok))
		    (lambda (keys . args)
		      (boot-format-fail)))
	     #f
	     )
    )
   (make <service>
     #:requires '(networking)
     ;; TODO: insert here mounting remote fs
     #:provides '(console-setup)
     #:start (lambda args
	       (boot-format "[....] Console setup...")
	       (when (file-exists? "/bin/setupcon")
		 (system* "/bin/setupcon" "--save"))
	       (boot-format-ok)
	       #t
	       )
     )
   (make <service>
     #:requires '(console-setup)
     #:provides '(login-helper)
     #:start (lambda args
	       (boot-format "[....] Registering boot in wtmp...")
	       ;; Create /var/run/utmp so we can login.
	       (close (open-output-file "/var/run/utmp"))
	       
	       (chmod "/var/run/utmp" #o664)
	       (let ((gid (catch-system-error (group:gid (getgrnam "utmp")))))
		  (when gid (chown "/var/run/utmp" -1 gid)))

	       (when (and (defined? 'wtmp-runlevel-set!) (file-exists? "/var/log/wtmp"))
		 (runlevel-set! "/var/run/utmp" "reboot" "~~" 0 BOOT_TIME "~")
		 (wtmp-runlevel-set! "/var/log/wtmp" "reboot" "~~" 0 BOOT_TIME "~")
		 (runlevel-set! "/var/run/utmp" "runlevel" "~~" (+ runlevel (* 256 (char->integer #\N))) RUN_LVL "~")
		 (wtmp-runlevel-set! "/var/log/wtmp" "runlevel" "~~" (+ runlevel (* 256 (char->integer #\N))) RUN_LVL "~")
		 )

	       (boot-format-ok)
	       #t
	       )
     #:stop (lambda (running)
	      (boot-format "[....] Registering shutdown in wtmp...")
	      ;; at that point, definition of wtmp-runlevel-set! is lost
	      ;; (because we are in module (guile-user)?), but stays executable
	      (when (and (defined? 'wtmp-runlevel-set! shepherd-module) (file-exists? "/var/log/wtmp"))
		(wtmp-runlevel-set! "/var/log/wtmp" "shutdown" "~~" 0 RUN_LVL "~~")
		)
	      (boot-format-ok)
	      #f
	      )
     )
   (make <service>
     #:requires '(login-helper)
     #:provides '(X11-common)
     #:start (lambda args
	       (boot-format "[....] Setting up X socket directories...")
	       (unless (file-exists? "/tmp/.X11-unix") (mkdir "/tmp/.X11-unix"))
	       (chmod "/tmp/.X11-unix" #o1777)
	       (unless (file-exists? "/tmp/.ICE-unix") (mkdir "/tmp/.ICE-unix"))
	       (chmod "/tmp/.ICE-unix" #o1777)
	       (boot-format-ok)
	       #t
	       )
     )
   (make <service>
     #:requires '(X11-common)
     #:provides '(single-user)
     #:start (lambda args
 	       (boot-format " -- Single user boot ok --\n")
	       #t
	       )
     #:stop (lambda (running)
	      ;; the console where the last steps of reboot/poweroff
	      ;; are recorded is in most cases /dev/tty1, so we go
	      ;; there
	      (if (file-exists? "/bin/chvt")
		  (system* "/bin/chvt" "1"))
 	      (boot-format " -- Rebooting or shutting down --\n")
	      #f
	      )
     )
   )

(catch 'system-error
       (lambda ()
	 (let ((services (scandir (cat %sysconfdir "/shepherd.d/packages") (lambda (x) (string-match ".*\\.scm$" x)))))
	   (for-each (lambda (x) 
		       (catch 'system-error
			      (lambda ()
				(load (cat %sysconfdir "/shepherd.d/packages/" x)))
			      (lambda (keys . args) #f)))
		     services))
	 )
       (lambda (keys . args) #f))

;; Services to start when shepherd starts; this is less verbose than (for-each-service start)

(define (for-each-service-start)
  (hash-for-each (lambda (name services)
                   (and=> ((@@ (shepherd service) lookup-canonical-service) name services)
                          (lambda (x) (unless (running? x) (start x)))))
                 (@@ (shepherd service) %services)))

;; no %services in shepherd 0.10 series
(if (defined? '%services (resolve-module '(shepherd service)))
    (for-each-service-start)
    (for-each-service start))

